/*
 * yasapi_resample.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
#define BUF_SIZE 8192

///////////////////////////////////////////////////////////////////////////////
// https://bug825153.bugzilla.mozilla.org/attachment.cgi?id=729339
#define MF_VERSION_VISTA    (0x0001<<16|MF_API_VERSION)
#define MF_VERSION_WIN7     (0x0002<<16|MF_API_VERSION)

#if 0 // {
static int debug=1;

///////////////////////////////////////////////////////////////////////////////
// https://code.google.com/p/bitspersampleconv2/wiki/HowToUseResamplerMFT
typedef struct Resample Resample;

struct Resample {
  IUnknown *pTransformUnk;
  IMFTransform *pTransform;
  IWMResamplerProps *pResamplerProps;
  IMFMediaType *pTypeInput;
  IMFMediaType *pTypeOutput;
  IMFMediaBuffer *pBuffer;
};

int ResampleCreate(Resample *pResample, WAVEFORMATEX *pwfxInput,
    WAVEFORMATEX *pwfxOutput);
void ResampleDestroy(Resample *pResample);

///////////////////////////////////////////////////////////////////////////////
int ResampleCreate(Resample *pResample, WAVEFORMATEX *pwfxInput,
    WAVEFORMATEX *pwfxOutput)
{
  IUnknown *pTransformUnk;
  IMFTransform *pTransform;
  IWMResamplerProps *pResamplerProps;
  IMFMediaType *pTypeInput;
  IMFMediaType *pTypeOutput;
  IMFMediaBuffer *pBuffer;
  HRESULT hr;

  // create a transform unknown.
  hr=CoCreateInstance(
    &CLSID_CResamplerMediaObject,   // _In_   REFCLSID rclsid,
    NULL,                           // _In_   LPUNKNOWN pUnkOuter,
    CLSCTX_INPROC_SERVER,           // _In_   DWORD dwClsContext,
    &IID_IUnknown,                  // _In_   REFIID riid,
    (LPVOID *)&pTransformUnk        // _Out_  LPVOID *ppv
  );

  if (FAILED(hr)) {
    DERROR(debug,REGDB_E_CLASSNOTREG,hr);
    DERROR(debug,CLASS_E_NOAGGREGATION,hr);
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating transform unknown");
    goto factory;
  }

  pResample->pTransformUnk=pTransformUnk;
  DPUTS(debug,"  transform unknown created\n");

  // create the transform.
  hr=pTransformUnk->lpVtbl->QueryInterface(pTransformUnk,
    &IID_IMFTransform,            // [in]   REFIID riid,
    (void **)&pTransform          // [out]  void **ppvObject
  );

  if (FAILED(hr)) {
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating transform");
    goto transform;
  }

  pResample->pTransform=pTransform;
  DPUTS(debug,"  transform created\n");

  // create resample properties.
  hr=pTransformUnk->lpVtbl->QueryInterface(pTransformUnk,
    &IID_IWMResamplerProps,       // [in]   REFIID riid,
    (void **)&pResamplerProps     // [out]  void **ppvObject
  );

  if (FAILED(hr)) {
    DERROR(debug,E_NOINTERFACE,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"creating resample properties");
    goto props;
  }

  pResample->pResamplerProps=pResamplerProps;
  DPUTS(debug,"  resample properties created\n");

  // best quality, always S_OK.
  hr=pResamplerProps->lpVtbl->SetHalfFilterLength(pResamplerProps,
    60                            // [in]  LONG lhalfFilterLen
  );

  // create the input type
  hr=MediaTypeCreateWFX(&pTypeInput,pwfxInput,"input",debug);

  if (FAILED(hr))
    goto input_cre;

  pTransform->lpVtbl->SetInputType(pTransform,
    0,                            // [in]  DWORD dwInputStreamID,
    pTypeInput,                   // [in]  IMFMediaType *pType,
    0                             // [in]  DWORD dwFlags
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_INVALIDMEDIATYPE,hr);
    DERROR(debug,MF_E_INVALIDSTREAMNUMBER,hr);
    DERROR(debug,MF_E_INVALIDTYPE,hr);
    DERROR(debug,MF_E_TRANSFORM_CANNOT_CHANGE_MEDIATYPE_WHILE_PROCESSING,hr);
    DERROR(debug,MF_E_TRANSFORM_TYPE_NOT_SET,hr);
    DERROR(debug,MF_E_UNSUPPORTED_D3D_TYPE,hr);
    DMESSAGE(debug,"setting input type");
    goto input_set;
  }

  pResample->pTypeInput=pTypeInput;
  DPUTS(debug,"  input type set\n");

  // create the output type
  hr=MediaTypeCreateWFX(&pTypeOutput,pwfxOutput,"output",debug);

  if (FAILED(hr))
    goto output_cre;

  pTransform->lpVtbl->SetOutputType(pTransform,
    0,                            // [in]  DWORD dwInputStreamID,
    pTypeOutput,                  // [in]  IMFMediaType *pType,
    0                             // [in]  DWORD dwFlags
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_INVALIDMEDIATYPE,hr);
    DERROR(debug,MF_E_INVALIDSTREAMNUMBER,hr);
    DERROR(debug,MF_E_INVALIDTYPE,hr);
    DERROR(debug,MF_E_TRANSFORM_CANNOT_CHANGE_MEDIATYPE_WHILE_PROCESSING,hr);
    DERROR(debug,MF_E_TRANSFORM_TYPE_NOT_SET,hr);
    DERROR(debug,MF_E_UNSUPPORTED_D3D_TYPE,hr);
    DMESSAGE(debug,"setting output type");
    goto output_set;
  }

  pResample->pTypeOutput=pTypeOutput;
  DPUTS(debug,"  output type set\n");

  // send the flush message.
  hr=pTransform->lpVtbl->ProcessMessage(pTransform,
    MFT_MESSAGE_COMMAND_FLUSH,  // [in]  MFT_MESSAGE_TYPE eMessage,
    0                           // [in]  ULONG_PTR ulParam
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_INVALIDSTREAMNUMBER,hr);
    DERROR(debug,MF_E_TRANSFORM_TYPE_NOT_SET,hr);
    DMESSAGE(debug,"sending flush message");
    goto message;
  }

  DPUTS(debug,"  flush message send\n");

  // send the notify start message.
  hr=pTransform->lpVtbl->ProcessMessage(pTransform,
    MFT_MESSAGE_NOTIFY_BEGIN_STREAMING,
                                // [in]  MFT_MESSAGE_TYPE eMessage,
    0                           // [in]  ULONG_PTR ulParam
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_INVALIDSTREAMNUMBER,hr);
    DERROR(debug,MF_E_TRANSFORM_TYPE_NOT_SET,hr);
    DMESSAGE(debug,"sending notify begin message");
    goto message;
  }

  DPUTS(debug,"  notify start begin send\n");

  // send the notify start message.
  hr=pTransform->lpVtbl->ProcessMessage(pTransform,
    MFT_MESSAGE_NOTIFY_START_OF_STREAM,
                                // [in]  MFT_MESSAGE_TYPE eMessage,
    0                           // [in]  ULONG_PTR ulParam
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_INVALIDSTREAMNUMBER,hr);
    DERROR(debug,MF_E_TRANSFORM_TYPE_NOT_SET,hr);
    DMESSAGE(debug,"sending notify start message");
    goto message;
  }

  DPUTS(debug,"  notify start message send\n");

  // create the memory buffer.
  hr=MFCreateMemoryBuffer(
    BUF_SIZE,                   // DWORD cbMaxLength,
    &pBuffer                    // IMFMediaBuffer **ppBuffer
  );
  
  if (FAILED(hr)) {
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"creating memory buffer");
    goto buffer;
  }

  pResample->pBuffer=pBuffer;
  DPUTS(debug,"  memory buffer created\n");

  return 0;
// cleanup:
  pBuffer->lpVtbl->Release(pBuffer);
buffer:
message:
output_set:
  pTypeOutput->lpVtbl->Release(pTypeOutput);
output_cre:
input_set:
  pTypeInput->lpVtbl->Release(pTypeInput);
input_cre:
  pResamplerProps->lpVtbl->Release(pResamplerProps);
props:
  pTransform->lpVtbl->Release(pTransform);
transform:
  pTransformUnk->lpVtbl->Release(pTransformUnk);
factory:
  return -1;
}

void ResampleDestroy(Resample *pResample)
{
  pResample->pBuffer->lpVtbl->Release(pResample->pBuffer);
  DPUTS(debug,"  memory buffer destroyed\n");
  pResample->pTypeOutput->lpVtbl->Release(pResample->pTypeOutput);
  DPUTS(debug,"  output type destroyed\n");
  pResample->pTypeInput->lpVtbl->Release(pResample->pTypeInput);
  DPUTS(debug,"  input type destroyed\n");
  pResample->pResamplerProps->lpVtbl->Release(pResample->pResamplerProps);
  DPUTS(debug,"  resample properties destroyed\n");
  pResample->pTransform->lpVtbl->Release(pResample->pTransform);
  DPUTS(debug,"  transform destroyed\n");
  pResample->pTransformUnk->lpVtbl->Release(pResample->pTransformUnk);
  DPUTS(debug,"  transform unknown destroyed\n");
}

int ResampleWrite(Resample *pResample, void *pData, DWORD cbMaxLength)
{
  IMFMediaBuffer *pBuffer;

  // create the memory buffer.
  hr=MFCreateMemoryBuffer(
    cbMaxLength,                // DWORD cbMaxLength,
    &pBuffer                    // IMFMediaBuffer **ppBuffer
  );
  
  if (FAILED(hr)) {
    DERROR(debug,E_OUTOFMEMORY,hr);
    DMESSAGE(debug,"creating memory buffer");
    goto buffer;
  }

  DPUTS(debug,"  memory buffer created\n");
// cleanup:
}

///////////////////////////////////////////////////////////////////////////////
void resample_hello()
{
  WAVEFORMATEXTENSIBLE wfxxInput;
  WAVEFORMATEXTENSIBLE wfxxOutput;
  Resample res;
  HRESULT hr;

  // starting up media foundation.
  hr=MFStartup(
    MF_VERSION_VISTA,               // ULONG Version,
    MFSTARTUP_NOSOCKET              // DWORD dwFlags = MFSTARTUP_FULL
  );

  if (FAILED(hr)) {
    DERROR(debug,MF_E_BAD_STARTUP_VERSION,hr);
    DERROR(debug,MF_E_DISABLED_IN_SAFEMODE,hr);
    DERROR(debug,E_NOTIMPL,hr);
    DMESSAGE(debug,"starting up media foundation");
    goto startup;
  }

  DPUTS(debug,"  media foundation started up\n");

  // set attributes.
  WaveFormatExtensibleSetup(&wfxxInput,NULL,44100,2,24);
  WaveFormatExtensibleSetup(&wfxxOutput,NULL,48000,2,24);

  // create the resampler.
  if (ResampleCreate(&res,&wfxxInput.Format,&wfxxOutput.Format)<0)
    goto res;

// cleanup:
  ResampleDestroy(&res);
res:
  MFShutdown();
  DPUTS(debug,"  media foundation shut down\n");
startup:
  ;
}
#else // } {
#if defined (YASAPI_DEBUG) // {
void resample_hello()
{
}
#endif // }
#endif // }
