/*
 * yasapi_timer.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

// http://www.virtualdub.org/blog/pivot/entry.php?id=272
int TimerCreate(Timer *pTimer, BOOL debug)
{
#if defined (YASAPI_DEBUG) // {
  pTimer->debug=debug;
#endif // }

  pTimer->hTimer=CreateWaitableTimer(
    NULL,           // _In_opt_  LPSECURITY_ATTRIBUTES lpTimerAttributes,
    FALSE,          // _In_      BOOL bManualReset,
    NULL            // _In_opt_  LPCTSTR lpTimerName
  );

  if (NULL==pTimer->hTimer) {
    DMESSAGE(debug,"creating waitable timer");
    goto timer;
  }

  return 0;
// cleanup:
  CloseHandle(pTimer->hTimer);
timer:
  return -1;
}

void TimerDestroy(Timer *pTimer)
{
  CloseHandle(pTimer->hTimer);
}

void TimerSet(Timer *pTimer, LONGLONG time)
{
  LARGE_INTEGER ui;

  // http://msdn.microsoft.com/en-us/library/windows/desktop/ms687012%28v=vs.85%29.aspx
  // Positive values indicate absolute time. 
  // Negative values indicate relative time.
  // in 100 nanosecond intervals
  ui.QuadPart=-time;

  SetWaitableTimer(
    pTimer->hTimer,   // _In_      HANDLE hTimer,
    &ui,              // _In_      const LARGE_INTEGER *pDueTime,
    0,                // _In_      LONG lPeriod,
    NULL,             // _In_opt_  PTIMERAPCROUTINE pfnCompletionRoutine,
    NULL,             // _In_opt_  LPVOID lpArgToCompletionRoutine,
    FALSE             // _In_      BOOL fResume
  );
}

#if 0 // {
void TimerSetMS(Timer *pTimer, DWORD ms)
{
  TimerSet(pTimer,10000ll*ms)
}
#endif // }

void TimerKill(Timer *pTimer)
{
  CancelWaitableTimer(pTimer->hTimer);
  TimerSet(pTimer,0);
}

void TimerSleep(Timer *pTimer, LONGLONG time)
{
  TimerSet(pTimer,time);
  WaitForSingleObject(pTimer->hTimer,INFINITE);
}

void TimerSleepTime(Timer *pTimer, LONGLONG time)
{
  TimerSleep(pTimer,10ll*time);
}

void TimerSleepMS(Timer *pTimer, LONGLONG ms)
{
  TimerSleep(pTimer,10000ll*ms);
}
